package com.tzld.piaoquan.recommend.server.implement;


import com.alibaba.fastjson.JSONObject;
import com.google.common.base.Stopwatch;
import com.google.common.reflect.TypeToken;
import com.tzld.piaoquan.recommend.server.common.base.RankItem;
import com.tzld.piaoquan.recommend.server.framework.candidiate.Candidate;
import com.tzld.piaoquan.recommend.server.framework.common.User;
import com.tzld.piaoquan.recommend.server.framework.merger.MergeUtils;
import com.tzld.piaoquan.recommend.server.framework.merger.StrategyQueue;
import com.tzld.piaoquan.recommend.server.framework.recaller.BaseRecaller;
import com.tzld.piaoquan.recommend.server.framework.recaller.provider.RedisBackedQueue;
import com.tzld.piaoquan.recommend.server.framework.score.ScorerUtils;
import com.tzld.piaoquan.recommend.server.framework.utils.RedisSmartClient;
import com.tzld.piaoquan.recommend.server.gen.recommend.RecommendRequest;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.rank.extractor.ExtractorUtils;
import com.tzld.piaoquan.recommend.server.service.rank.extractor.RankExtractorItemFeature;
import com.tzld.piaoquan.recommend.server.service.rank.extractor.RankExtractorUserFeature;
import com.tzld.piaoquan.recommend.server.util.CommonCollectionUtils;
import com.tzld.piaoquan.recommend.server.util.JSONUtils;
import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.redis.connection.RedisConnectionFactory;
import org.springframework.data.redis.connection.RedisStandaloneConfiguration;
import org.springframework.data.redis.connection.jedis.JedisConnectionFactory;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.serializer.StringRedisSerializer;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.stream.Collectors;

@Service
public class TopRecommendPipeline {

    private static final Logger log = LoggerFactory.getLogger(TopRecommendPipeline.class);

    public static final String MERGE_CONF = "merge_config.conf";

    @Resource
    private RedisSmartClient client;
    @Resource
    public RedisTemplate<String, String> redisTemplate;
    private RedisBackedQueue queueProvider;

    @PostConstruct
    public void init() {
        queueProvider = new RedisBackedQueue(client, 15 * 60 * 1000L);
    }

    public List<Video> feeds(final RecommendRequest requestData,
                             final int requestIndex,
                             final User userInfo, Boolean logPrint) {
        // Step 1: Attention extraction
        Stopwatch stopwatch = Stopwatch.createStarted();
        stopwatch.reset().start();
        List<RankItem> rankItems = feedByRec(requestData, requestIndex, userInfo, logPrint);
        if (rankItems == null || rankItems.isEmpty()) {
            return new ArrayList<>();
        }
        if (logPrint) {
            log.info("traceId = {}, cost = {}, feeds rankItems = {}", requestData.getRequestId(),
                    stopwatch.elapsed().toMillis(), JSONUtils.toJson(rankItems));
        }
        stopwatch.reset().start();
        List<Video> videos = rankItem2Video(rankItems);
        if (logPrint) {
            log.info("traceId = {}, cost = {}, videos = {}", requestData.getRequestId(),
                    stopwatch.elapsed().toMillis(), JSONUtils.toJson(videos));
        }
        return videos;
    }

    public List<Double> getStaticData(Map<String, Map<String, Double>> itemRealMap,
                                      List<String> datehours, String key){
        List<Double> views = new LinkedList<>();
        Map<String, Double> tmp = itemRealMap.getOrDefault(key, new HashMap<>());
        for (String dh : datehours){
            views.add(tmp.getOrDefault(dh, 0.0D) +
                    (views.isEmpty() ? 0.0: views.get(views.size()-1))
            );
        }
        return views;
    }

    public Double calScoreWeight(List<Double> data){
        Double up = 0.0;
        Double down = 0.0;
        for (int i=0; i<data.size(); ++i){
            up += 1.0 / (i + 1) * data.get(i);
            down += 1.0 / (i + 1);
        }
        return down > 1E-8? up / down: 0.0;
    }
    public List<Double> getRateData(List<Double> ups, List<Double> downs, Double up, Double down){
        List<Double> data = new LinkedList<>();
        for(int i=0; i<ups.size(); ++i){
            data.add(
                    (ups.get(i) + up) / (downs.get(i) + down)
            );
        }
        return data;
    }

    public List<RankItem> feedByRec(final RecommendRequest requestData,
                                    final int requestIndex,
                                    final User userInfo, Boolean logPrint) {
        int recallNum = 150;

        Stopwatch stopwatch = Stopwatch.createStarted();

        // Step 2: create top queue
        stopwatch.reset().start();
        StrategyQueue topQueue = MergeUtils.createTopQueue(MERGE_CONF, "top-queue");
        if (logPrint) {
            log.info("traceId = {}, cost = {}, topQueue = {}", requestData.getRequestId(),
                    stopwatch.elapsed().toMillis(), JSONUtils.toJson(topQueue));
        }

        // Step 3: Candidate
        stopwatch.reset().start();
        Map<String, Candidate> candidates = new HashMap<String, Candidate>();
        topQueue.candidate(candidates, recallNum, userInfo, requestData, 0, 0);
        if (logPrint) {
            log.info("traceId = {}, cost = {}, candidates = {}", requestData.getRequestId(),
                    stopwatch.elapsed().toMillis(), JSONUtils.toJson(candidates));
        }


        // Step 4: Recalling & Basic Scoring
        stopwatch.reset().start();
        BaseRecaller recaller = new BaseRecaller(queueProvider);
        List<RankItem> items = recaller.recalling(requestData, userInfo, requestIndex, new ArrayList<>(candidates.values()));
        if (logPrint) {
            log.info("traceId = {}, cost = {}, items = {}", requestData.getRequestId(),
                    stopwatch.elapsed().toMillis(), JSONUtils.toJson(items));
        }


        // Step 4: Advance Scoring
//        timestamp = System.currentTimeMillis();
//        ScorerPipeline scorerPipeline = getScorerPipeline(requestData);
//        items = scorerPipeline.scoring(requestData, userInfo, requestIndex, items);
        if (CollectionUtils.isEmpty(items)) {
            return new ArrayList<>();
        }

        stopwatch.reset().start();
        // Step 5: Merger
        MergeUtils.distributeItemsToMultiQueues(topQueue, items);
        topQueue.merge(recallNum * 3, userInfo, requestData, requestIndex, 0);

        // 多样性融合
        List<RankItem> mergeItems = topQueue.getItems();
        if (CollectionUtils.isEmpty(mergeItems)) {
            return new ArrayList<>();
        }
        duplicate(mergeItems);

        if (logPrint) {
            log.info("traceId = {}, cost = {}, mergeItems = {}", requestData.getRequestId(),
                    stopwatch.elapsed().toMillis(), JSONUtils.toJson(mergeItems));
        }
//        MergeUtils.diversityRerank(mergeItems, SimilarityUtils.getIsSameUserTagOrCategoryFunc(), recallNum, 6, 2);

        // Step 6: Global Rank & subList
        // TODO 前置和后置处理逻辑 hardcode，后续优化
        stopwatch.reset().start();
        List<RankItem> rovRecallRankNewScore = rankByScore(mergeItems, requestData);
        if (logPrint) {
            log.info("traceId = {}, cost = {}, rovRecallRankNewScore = {}", requestData.getRequestId(),
                    stopwatch.elapsed().toMillis(), JSONUtils.toJson(rovRecallRankNewScore));
        }

        return rovRecallRankNewScore;
    }

    private List<Video> rankItem2Video(List<RankItem> items) {
        List<String> rtFeaPartKey = new ArrayList<>(Arrays.asList("item_rt_fea_1day_partition", "item_rt_fea_1h_partition"));
        List<String> rtFeaPartKeyResult = this.redisTemplate.opsForValue().multiGet(rtFeaPartKey);
        Calendar calendar = Calendar.getInstance();
        String date = new SimpleDateFormat("yyyyMMdd").format(calendar.getTime());
        String hour = new SimpleDateFormat("HH").format(calendar.getTime());
        String rtFeaPart1h = date + hour;
        if (rtFeaPartKeyResult != null){
            if (rtFeaPartKeyResult.get(1) != null){
                rtFeaPart1h = rtFeaPartKeyResult.get(1);
            }
        }
        // 2 统计分
        String cur = rtFeaPart1h;
        List<String> datehours = new LinkedList<>();
        for (int i=0; i<24; ++i){
            datehours.add(cur);
            cur = ExtractorUtils.subtractHours(cur, 1);
        }
        for (RankItem item : items){
            Map<String, String> itemBasicMap = item.getItemBasicFeature();
            Map<String, Map<String, Double>> itemRealMap = item.getItemRealTimeFeature();
            List<Double> views = getStaticData(itemRealMap, datehours, "view_pv_list_1h");
            List<Double> plays = getStaticData(itemRealMap, datehours, "play_pv_list_1h");
            List<Double> shares = getStaticData(itemRealMap, datehours, "share_pv_list_1h");
            List<Double> returns = getStaticData(itemRealMap, datehours, "p_return_uv_list_1h");
            List<Double> allreturns = getStaticData(itemRealMap, datehours, "return_uv_list_1h");

            List<Double> share2return = getRateData(returns, shares, 1.0, 1000.0);
            Double share2returnScore = calScoreWeight(share2return);
            List<Double> view2return = getRateData(returns, views, 1.0, 1000.0);
            Double view2returnScore = calScoreWeight(view2return);
            List<Double> view2play = getRateData(plays, views, 1.0, 1000.0);
            Double view2playScore = calScoreWeight(view2play);
            List<Double> play2share = getRateData(shares, plays, 1.0, 1000.0);
            Double play2shareScore = calScoreWeight(play2share);
            item.scoresMap.put("share2returnScore", share2returnScore);
            item.scoresMap.put("view2returnScore", view2returnScore);
            item.scoresMap.put("view2playScore", view2playScore);
            item.scoresMap.put("play2shareScore", play2shareScore);

            Double allreturnsScore = calScoreWeight(allreturns);
            item.scoresMap.put("allreturnsScore", allreturnsScore);
        }
        // 3 融合公式
        List<Video> result = new ArrayList<>();
        for (RankItem item : items){
            double score = item.getScoreStr() *
                    item.scoresMap.getOrDefault("share2returnScore", 0.0) *
                    Math.log(1 + item.scoresMap.getOrDefault("allreturnsScore", 0.0));
            Video video = new Video();
            video.setVideoId(Long.parseLong(item.getId()));
            video.setPushFrom(item.getQueue());
            video.setScore(score);
            video.setSortScore(score);
            video.setScoreStr(item.getScoreStr());
            video.setScoresMap(item.getScoresMap());
            result.add(video);
        }
        Collections.sort(result, Comparator.comparingDouble(o -> -o.getSortScore()));
        return result;
    }

    private void duplicate(List<RankItem> items) {
        Set<String> ids = new HashSet<>();
        List<RankItem> result = new ArrayList<>();
        for (RankItem item : items) {
            if (ids.contains(item.getId())) {
                continue;
            }
            ids.add(item.getId());
            result.add(item);
        }
        items.clear();
        items.addAll(result);
    }

    public List<RankItem> rankByScore(List<RankItem> rankItems, RecommendRequest param){
        List<RankItem> result = new ArrayList<>();
        if (rankItems.isEmpty()){
            return result;
        }

        RedisStandaloneConfiguration redisSC = new RedisStandaloneConfiguration();
        redisSC.setPort(6379);
        redisSC.setPassword("Wqsd@2019");
        redisSC.setHostName("r-bp1pi8wyv6lzvgjy5z.redis.rds.aliyuncs.com");
        RedisConnectionFactory connectionFactory = new JedisConnectionFactory(redisSC);
        RedisTemplate<String, String> redisTemplate = new RedisTemplate<>();
        redisTemplate.setConnectionFactory(connectionFactory);
        redisTemplate.setDefaultSerializer(new StringRedisSerializer());
        redisTemplate.afterPropertiesSet();

        // 0: 场景特征处理
        Map<String, String> sceneFeatureMap =  this.getSceneFeature(param);

        // 1: user特征处理
        Map<String, String> userFeatureMap = new HashMap<>();
        if (param.getMid() != null && !param.getMid().isEmpty()){
            String midKey = "user_info_4video_" + param.getMid();
            String userFeatureStr = redisTemplate.opsForValue().get(midKey);
            if (userFeatureStr != null){
                try{
                    userFeatureMap = JSONUtils.fromJson(userFeatureStr,
                            new TypeToken<Map<String, String>>() {},
                            userFeatureMap);
                }catch (Exception e){
                    log.error(String.format("parse user json is wrong in {} with {}", this.getClass().getSimpleName(), e));
                }
            }else{
                JSONObject obj = new JSONObject();
                obj.put("name", "user_key_in_model_is_null");
                obj.put("class", this.getClass().getSimpleName());
                log.info(obj.toString());
//                return videos;
            }
        }
        final Set<String> userFeatureSet = new HashSet<>(Arrays.asList(
                "machineinfo_brand", "machineinfo_model", "machineinfo_platform", "machineinfo_system",
                "u_1day_exp_cnt", "u_1day_click_cnt", "u_1day_share_cnt", "u_1day_return_cnt",
                "u_3day_exp_cnt", "u_3day_click_cnt", "u_3day_share_cnt", "u_3day_return_cnt"
        ));
        Iterator<Map.Entry<String, String>> iterator = userFeatureMap.entrySet().iterator();
        while (iterator.hasNext()) {
            Map.Entry<String, String> entry = iterator.next();
            if (!userFeatureSet.contains(entry.getKey())) {
                iterator.remove();
            }
        }
        Map<String, String> f1 = RankExtractorUserFeature.getOriginFeature(userFeatureMap,
                new HashSet<String>(Arrays.asList(
                        "machineinfo_brand", "machineinfo_model", "machineinfo_platform", "machineinfo_system"
                ))
        );
        Map<String, String> f2 = RankExtractorUserFeature.getUserRateFeature(userFeatureMap);
        Map<String, String> f3 = RankExtractorUserFeature.cntFeatureChange(userFeatureMap,
                new HashSet<String>(Arrays.asList(
                        "u_1day_exp_cnt", "u_1day_click_cnt", "u_1day_share_cnt", "u_1day_return_cnt",
                        "u_3day_exp_cnt", "u_3day_click_cnt", "u_3day_share_cnt", "u_3day_return_cnt"
                ))
        );
        f1.putAll(f2);
        f1.putAll(f3);
        log.info("userFeature in model = {}", JSONUtils.toJson(f1));

        // 2-1: item特征处理
        final Set<String> itemFeatureSet = new HashSet<>(Arrays.asList(
                "total_time", "play_count_total",
                "i_1day_exp_cnt", "i_1day_click_cnt", "i_1day_share_cnt", "i_1day_return_cnt",
                "i_3day_exp_cnt", "i_3day_click_cnt", "i_3day_share_cnt", "i_3day_return_cnt"
        ));

        List<String> videoIds = CommonCollectionUtils.toListDistinct(rankItems, RankItem::getId);
        List<String> videoFeatureKeys = videoIds.stream().map(r-> "video_info_" + r)
                .collect(Collectors.toList());
        List<String> videoFeatures = redisTemplate.opsForValue().multiGet(videoFeatureKeys);
        if (videoFeatures != null){
            for (int i=0; i<videoFeatures.size(); ++i){
                String vF = videoFeatures.get(i);
                Map<String, String> vfMap = new HashMap<>();
                if (vF == null){
                    continue;
                }
                try{
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {}, vfMap);
                    rankItems.get(i).setItemBasicFeature(vfMap);
                    Iterator<Map.Entry<String, String>> iteratorIn = vfMap.entrySet().iterator();
                    while (iteratorIn.hasNext()) {
                        Map.Entry<String, String> entry = iteratorIn.next();
                        if (!itemFeatureSet.contains(entry.getKey())) {
                            iteratorIn.remove();
                        }
                    }
                    Map<String, String> f4 = RankExtractorItemFeature.getItemRateFeature(vfMap);
                    Map<String, String> f5 = RankExtractorItemFeature.cntFeatureChange(vfMap,
                            new HashSet<>(Arrays.asList(
                                    "total_time", "play_count_total",
                                    "i_1day_exp_cnt", "i_1day_click_cnt", "i_1day_share_cnt", "i_1day_return_cnt",
                                    "i_3day_exp_cnt", "i_3day_click_cnt", "i_3day_share_cnt", "i_3day_return_cnt"))
                    );
                    f4.putAll(f5);
                    rankItems.get(i).setFeatureMap(f4);
                }catch (Exception e){
                    log.error(String.format("parse video json is wrong in {} with {}", this.getClass().getSimpleName(), e));
                }
            }
        }
        // 2-2: item 实时特征处理
        List<String> rtFeaPartKey = new ArrayList<>(Arrays.asList("item_rt_fea_1day_partition", "item_rt_fea_1h_partition"));
        List<String> rtFeaPartKeyResult = this.redisTemplate.opsForValue().multiGet(rtFeaPartKey);
        Calendar calendar = Calendar.getInstance();
        String date = new SimpleDateFormat("yyyyMMdd").format(calendar.getTime());
        String hour = new SimpleDateFormat("HH").format(calendar.getTime());
        String rtFeaPart1day = date + hour;
        String rtFeaPart1h = date + hour;
        if (rtFeaPartKeyResult != null){
            if (rtFeaPartKeyResult.get(0) != null){
                rtFeaPart1day = rtFeaPartKeyResult.get(0);
            }
            if (rtFeaPartKeyResult.get(1) != null){
                rtFeaPart1h = rtFeaPartKeyResult.get(1);
            }
        }

        List<String> videoRtKeys1 = videoIds.stream().map(r-> "item_rt_fea_1day_" + r)
                .collect(Collectors.toList());
        List<String> videoRtKeys2 = videoIds.stream().map(r-> "item_rt_fea_1h_" + r)
                .collect(Collectors.toList());
        videoRtKeys1.addAll(videoRtKeys2);
        List<String> videoRtFeatures = this.redisTemplate.opsForValue().multiGet(videoRtKeys1);


        if (videoRtFeatures != null){
            int j = 0;
            for (RankItem item: rankItems){
                String vF = videoRtFeatures.get(j);
                ++j;
                if (vF == null){
                    continue;
                }
                Map<String, String> vfMap = new HashMap<>();
                Map<String, Map<String, Double>> vfMapNew = new HashMap<>();
                try{
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {}, vfMap);
                    for (Map.Entry<String, String> entry : vfMap.entrySet()){
                        String value = entry.getValue();
                        if (value == null){
                            continue;
                        }
                        String [] var1 = value.split(",");
                        Map<String, Double> tmp = new HashMap<>();
                        for (String var2 : var1){
                            String [] var3 = var2.split(":");
                            tmp.put(var3[0], Double.valueOf(var3[1]));
                        }
                        vfMapNew.put(entry.getKey(), tmp);
                    }
                }catch (Exception e){
                    log.error(String.format("parse video item_rt_fea_1day_ json is wrong in {} with {}",
                            this.getClass().getSimpleName(), e));
                }
                Map<String, String> f8 = RankExtractorItemFeature.getItemRealtimeRate(vfMapNew, rtFeaPart1day);
                item.getFeatureMap().putAll(f8);
            }
            for (RankItem item: rankItems){
                String vF = videoRtFeatures.get(j);
                ++j;
                if (vF == null){
                    continue;
                }
                Map<String, String> vfMap = new HashMap<>();
                Map<String, Map<String, Double>> vfMapNew = new HashMap<>();
                try{
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {}, vfMap);

                    for (Map.Entry<String, String> entry : vfMap.entrySet()){
                        String value = entry.getValue();
                        if (value == null){
                            continue;
                        }
                        String [] var1 = value.split(",");
                        Map<String, Double> tmp = new HashMap<>();
                        for (String var2 : var1){
                            String [] var3 = var2.split(":");
                            tmp.put(var3[0], Double.valueOf(var3[1]));
                        }
                        vfMapNew.put(entry.getKey(), tmp);
                    }
                    item.setItemRealTimeFeature(vfMapNew);
                }catch (Exception e){
                    log.error(String.format("parse video item_rt_fea_1h_ json is wrong in {} with {}",
                            this.getClass().getSimpleName(), e));
                }
                Map<String, String> f8 = RankExtractorItemFeature.getItemRealtimeRate(vfMapNew, rtFeaPart1h);
                item.getFeatureMap().putAll(f8);
            }
        }


        log.info("ItemFeature = {}", JSONUtils.toJson(videoFeatures));


        List<RankItem> rovRecallScore = ScorerUtils.getScorerPipeline(ScorerUtils.BASE_CONF_FEED)
                .scoring(sceneFeatureMap, userFeatureMap, rankItems);
        log.info("mergeAndRankRovRecallNew rovRecallScore={}", JSONUtils.toJson(rovRecallScore));
        JSONObject obj = new JSONObject();
        obj.put("name", "user_key_in_model_is_not_null");
        obj.put("class", this.getClass().getSimpleName());
        log.info(obj.toString());
        return rovRecallScore;
    }

    private Map<String, String> getUserFeatureMap(RecommendRequest param, List<RankItem> rankItems) {
        Map<String, String> userFeatureMap = new HashMap<>(64);
        if (param.getMid() != null && !param.getMid().isEmpty()){
            String midKey = "user_info_4video_" + param.getMid();
            String userFeatureStr = redisTemplate.opsForValue().get(midKey);
            if (userFeatureStr != null){
                try{
                    userFeatureMap = JSONUtils.fromJson(userFeatureStr,
                            new TypeToken<Map<String, String>>() {},
                            userFeatureMap);
                }catch (Exception e){
                    log.error(String.format("parse user json is wrong in {} with {}", this.getClass().getSimpleName(), e));
                }
            }else{
                JSONObject obj = new JSONObject();
                obj.put("name", "user_key_in_model_is_null");
                obj.put("class", this.getClass().getSimpleName());
                log.info(obj.toString());
//                return videos;
            }
        }
        final Set<String> userFeatureSet = new HashSet<>(Arrays.asList(
                "machineinfo_brand", "machineinfo_model", "machineinfo_platform", "machineinfo_system",
                "u_1day_exp_cnt", "u_1day_click_cnt", "u_1day_share_cnt", "u_1day_return_cnt",
                "u_3day_exp_cnt", "u_3day_click_cnt", "u_3day_share_cnt", "u_3day_return_cnt"
        ));
        Iterator<Map.Entry<String, String>> iterator = userFeatureMap.entrySet().iterator();
        while (iterator.hasNext()) {
            Map.Entry<String, String> entry = iterator.next();
            if (!userFeatureSet.contains(entry.getKey())) {
                iterator.remove();
            }
        }
        Map<String, String> f1 = RankExtractorUserFeature.getOriginFeature(userFeatureMap,
                new HashSet<String>(Arrays.asList(
                        "machineinfo_brand", "machineinfo_model", "machineinfo_platform", "machineinfo_system"
                ))
        );
        Map<String, String> f2 = RankExtractorUserFeature.getUserRateFeature(userFeatureMap);
        Map<String, String> f3 = RankExtractorUserFeature.cntFeatureChange(userFeatureMap,
                new HashSet<String>(Arrays.asList(
                        "u_1day_exp_cnt", "u_1day_click_cnt", "u_1day_share_cnt", "u_1day_return_cnt",
                        "u_3day_exp_cnt", "u_3day_click_cnt", "u_3day_share_cnt", "u_3day_return_cnt"
                ))
        );
        f1.putAll(f2);
        f1.putAll(f3);
        log.info("userFeature in model = {}", JSONUtils.toJson(f1));

        // 2-1: item特征处理
        final Set<String> itemFeatureSet = new HashSet<>(Arrays.asList(
                "total_time", "play_count_total",
                "i_1day_exp_cnt", "i_1day_click_cnt", "i_1day_share_cnt", "i_1day_return_cnt",
                "i_3day_exp_cnt", "i_3day_click_cnt", "i_3day_share_cnt", "i_3day_return_cnt"
        ));

        List<Long> videoIds = CommonCollectionUtils.toListDistinct(rankItems, RankItem::getVideoId);
        List<String> videoFeatureKeys = videoIds.stream().map(r-> "video_info_" + r)
                .collect(Collectors.toList());
        List<String> videoFeatures = redisTemplate.opsForValue().multiGet(videoFeatureKeys);
        if (videoFeatures != null){
            for (int i=0; i<videoFeatures.size(); ++i){
                String vF = videoFeatures.get(i);
                Map<String, String> vfMap = new HashMap<>();
                if (vF == null){
                    continue;
                }
                try{
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {}, vfMap);
                    Map<String, String> vfMapCopy = new HashMap<>(vfMap);
                    rankItems.get(i).setItemBasicFeature(vfMapCopy);
                    Iterator<Map.Entry<String, String>> iteratorIn = vfMap.entrySet().iterator();
                    while (iteratorIn.hasNext()) {
                        Map.Entry<String, String> entry = iteratorIn.next();
                        if (!itemFeatureSet.contains(entry.getKey())) {
                            iteratorIn.remove();
                        }
                    }
                    Map<String, String> f4 = RankExtractorItemFeature.getItemRateFeature(vfMap);
                    Map<String, String> f5 = RankExtractorItemFeature.cntFeatureChange(vfMap,
                            new HashSet<String>(Arrays.asList(
                                    "total_time", "play_count_total",
                                    "i_1day_exp_cnt", "i_1day_click_cnt", "i_1day_share_cnt", "i_1day_return_cnt",
                                    "i_3day_exp_cnt", "i_3day_click_cnt", "i_3day_share_cnt", "i_3day_return_cnt"))
                    );
                    f4.putAll(f5);
                    rankItems.get(i).setFeatureMap(f4);
                }catch (Exception e){
                    log.error(String.format("parse video json is wrong in {} with {}", this.getClass().getSimpleName(), e));
                }
            }
        }
        // 2-2: item 实时特征处理
        List<String> rtFeaPartKey = new ArrayList<>(Arrays.asList("item_rt_fea_1day_partition", "item_rt_fea_1h_partition"));
        List<String> rtFeaPartKeyResult = this.redisTemplate.opsForValue().multiGet(rtFeaPartKey);
        Calendar calendar = Calendar.getInstance();
        String date = new SimpleDateFormat("yyyyMMdd").format(calendar.getTime());
        String hour = new SimpleDateFormat("HH").format(calendar.getTime());
        String rtFeaPart1day = date + hour;
        String rtFeaPart1h = date + hour;
        if (rtFeaPartKeyResult != null){
            if (rtFeaPartKeyResult.get(0) != null){
                rtFeaPart1day = rtFeaPartKeyResult.get(0);
            }
            if (rtFeaPartKeyResult.get(1) != null){
                rtFeaPart1h = rtFeaPartKeyResult.get(1);
            }
        }

        List<String> videoRtKeys1 = videoIds.stream().map(r-> "item_rt_fea_1day_" + r)
                .collect(Collectors.toList());
        List<String> videoRtKeys2 = videoIds.stream().map(r-> "item_rt_fea_1h_" + r)
                .collect(Collectors.toList());
        videoRtKeys1.addAll(videoRtKeys2);
        List<String> videoRtFeatures = this.redisTemplate.opsForValue().multiGet(videoRtKeys1);


        if (videoRtFeatures != null){
            int j = 0;
            for (RankItem item: rankItems){
                ++j;
                if (j >= rankItems.size()) {
                    continue;
                }
                Map<String, String> vfMap = new HashMap<>();
                Map<String, Map<String, Double>> vfMapNew = new HashMap<>();
                try {
                    String vF = videoRtFeatures.get(j);
                    if (vF == null) {
                        continue;
                    }
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {
                    }, vfMap);
                    for (Map.Entry<String, String> entry : vfMap.entrySet()) {
                        String value = entry.getValue();
                        if (value == null) {
                            continue;
                        }
                        String[] var1 = value.split(",");
                        Map<String, Double> tmp = new HashMap<>();
                        for (String var2 : var1) {
                            String[] var3 = var2.split(":");
                            tmp.put(var3[0], Double.valueOf(var3[1]));
                        }
                        vfMapNew.put(entry.getKey(), tmp);
                    }
                }catch (Exception e){
                    log.error(String.format("parse video item_rt_fea_1day_ json is wrong in {} with {}",
                            this.getClass().getSimpleName(), e));
                }
                Map<String, String> f8 = RankExtractorItemFeature.getItemRealtimeRate(vfMapNew, rtFeaPart1day);
                item.getFeatureMap().putAll(f8);
            }
            for (RankItem item: rankItems){
                ++j;
                if (j >= rankItems.size()) {
                    continue;
                }
                Map<String, String> vfMap = new HashMap<>();
                Map<String, Map<String, Double>> vfMapNew = new HashMap<>();
                try {
                    String vF = videoRtFeatures.get(j);
                    if (vF == null) {
                        continue;
                    }
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {
                    }, vfMap);

                    for (Map.Entry<String, String> entry : vfMap.entrySet()) {
                        String value = entry.getValue();
                        if (value == null) {
                            continue;
                        }
                        String[] var1 = value.split(",");
                        Map<String, Double> tmp = new HashMap<>();
                        for (String var2 : var1) {
                            String [] var3 = var2.split(":");
                            tmp.put(var3[0], Double.valueOf(var3[1]));
                        }
                        vfMapNew.put(entry.getKey(), tmp);
                    }
                    item.setItemRealTimeFeature(vfMapNew);
                }catch (Exception e){
                    log.error(String.format("parse video item_rt_fea_1h_ json is wrong in {} with {}",
                            this.getClass().getSimpleName(), e));
                }
                Map<String, String> f8 = RankExtractorItemFeature.getItemRealtimeRate(vfMapNew, rtFeaPart1h);
                item.getFeatureMap().putAll(f8);
            }
        }


        log.info("ItemFeature = {}", JSONUtils.toJson(videoFeatures));
        return userFeatureMap;
    }

    private Map<String, String> getSceneFeature(RecommendRequest param) {
        Map<String, String> sceneFeatureMap = new HashMap<>();
        String provinceCn = param.getProvince();
        provinceCn = provinceCn.replaceAll("省$", "");
        sceneFeatureMap.put("ctx_region", provinceCn);
        String city = param.getCity();
        if ("台北市".equals(city) |
                "高雄市".equals(city) |
                "台中市".equals(city) |
                "桃园市".equals(city) |
                "新北市".equals(city) |
                "台南市".equals(city) |
                "基隆市".equals(city) |
                "吉林市".equals(city) |
                "新竹市".equals(city) |
                "嘉义市".equals(city)
        ){
        }else{
            city = city.replaceAll("市$", "");
        }
        sceneFeatureMap.put("ctx_city", city);

        Calendar calendar = Calendar.getInstance();
        sceneFeatureMap.put("ctx_week", (calendar.get(Calendar.DAY_OF_WEEK) + 6) % 7 + "");
        sceneFeatureMap.put("ctx_hour", new SimpleDateFormat("HH").format(calendar.getTime()));

        return sceneFeatureMap;
    }




}
